import random
from phenotype import Phenotype
from genotype import Genotype
import math
from operator import attrgetter
from Enums import *
from CsoundHandler import *

# The class containing all genetic operators and methods
# It may contain some methods that are not classified as genetic operators, but they were just placed here
# for convenient reasons

"""
Spike Train Distance Metrics (SDM)
We will use three basic SDMs:
- Spike Time Distance Metric
- Spike Interval Distance Metric
- Waveform Distance Metric

For the two first spike metrics, the train must be preprocessed to define when
it is spikeing. This is done by these three steps:
1. The activation threshold (val_t) is when the neuron is concidered to be spiking.
We define val_t to be 0 mV.
2. Move a k-timestep window (e.g. k = 5) along the spike train, and any activation value that is a) in
the exact middle of the time window, b) above T, and c) the maximum of all activations in the time
window, is considered a spike.
3. This window will prevent double-counting spikes; since each spike will often have a few
points above the threshold.

This algorithm will generate a list of spike times to the given spike train,
where the former simply lists the time points at which spikes occur,
while the latter lists the activation levels for every time point of the simulation.
"""


# Transforming genoes into phenoes by converting the binary data into valid float parameter values
def genoToPheno(genotypes):
    phenotypes = []
    for i in range(0, len(genotypes)):
        geno = genotypes[i]
        pheno = Phenotype()

        for j in range(len(geno.order)):
            # Distortion value
            if geno.order[j] == Enums.DISTORTION:
                intValue = float(binaryToInteger(geno.val_distortion)) * (52.0 / 64.0)
                pheno.parameters.append(round((float(intValue) / 10.0), 1) + 0.9)
                continue

            # Flanger value
            if geno.order[j] == Enums.FLANGER:
                intValue = float(binaryToInteger(geno.val_flanger)) * (102.0 / 128.0)
                pheno.parameters.append(round((float(intValue) / 1000.0), 3) - 0.001)

            # Tremolo value
            if geno.order[j] == Enums.TREMOLO:
                intValue = float(binaryToInteger(geno.val_tremolo1)) * (12.0 / 16.0)
                pheno.parameters.append(round((float(intValue) / 10.0), 1) - 0.1)
                intValue = float(binaryToInteger(geno.val_tremolo2)) * (12.0 / 16.0)
                pheno.parameters.append(round((float(intValue) / 10.0), 1) - 0.1)

            # Chorus value
            if geno.order[j] == Enums.CHORUS:
                intValue = float(binaryToInteger(geno.val_chorus1)) * (12.0 / 16.0)
                pheno.parameters.append(round((float(intValue) / 10.0), 1) - 0.1)
                intValue = float(binaryToInteger(geno.val_chorus2)) * (12.0 / 16.0)
                pheno.parameters.append(round((float(intValue) / 10.0), 1) - 0.1)

        # Genes
        pheno.data.extend(geno.data)
        pheno.order.extend(geno.order)

        phenotypes.append(pheno)

    return phenotypes

# Applies effects and assigns a fitness value by using the distance metrics in CsoundHandler
def testSound(pheno, targetFilepath, fftlength):
    tempParameters = list(pheno.parameters)
    firstEffectFinished = False
    success = True
    skip = False
    if sum(tempParameters) == 0:
        success = applyZero("cnoteshort.wav")
        skip = True
    if not skip:
        for j in range(len(pheno.order)):
            if success:
                if pheno.order[j] == Enums.DISTORTION:
                    if tempParameters[0] < 1.0:
                        del tempParameters[0]
                        success = True
                        continue
                    if firstEffectFinished:
                        success = applyDistortion("copy.wav", "sound.wav", tempParameters[0])
                        del tempParameters[0]
                        continue
                    else:
                        success = applyDistortion("ctoneshort.wav", "sound.wav", tempParameters[0])
                        del tempParameters[0]
                        firstEffectFinished = True
                        continue
                if pheno.order[j] == Enums.FLANGER:
                    if tempParameters[0] < 0:
                        del tempParameters[0]
                        success = True
                        continue
                    if firstEffectFinished:
                        success = applyFlanger("copy.wav", "sound.wav", tempParameters[0])
                        del tempParameters[0]
                        continue
                    else:
                        success = applyFlanger("ctoneshort.wav", "sound.wav", tempParameters[0])
                        del tempParameters[0]
                        firstEffectFinished = True
                        continue
                if pheno.order[j] == Enums.TREMOLO:
                    if tempParameters[0] < 0 and tempParameters[1] < 0:
                        del tempParameters[0]
                        del tempParameters[0]
                        success = True
                        continue
                    if tempParameters[0] < 0:
                        tempParameters[0] = 0
                    if tempParameters[1] < 0:
                        tempParameters[1] = 0
                    if firstEffectFinished:
                        success = applyTremolo("copy.wav", "sound.wav", tempParameters[0], tempParameters[1])
                        del tempParameters[0]
                        del tempParameters[0]
                        continue
                    else:
                        success = applyTremolo("ctoneshort.wav", "sound.wav", tempParameters[0], tempParameters[1])
                        del tempParameters[0]
                        del tempParameters[0]
                        firstEffectFinished = True
                        continue
                if pheno.order[j] == Enums.CHORUS:
                    if tempParameters[0] < 0 and tempParameters[1] < 0:
                        del tempParameters[0]
                        del tempParameters[0]
                        success = True
                        continue
                    if tempParameters[0] < 0:
                        tempParameters[0] = 0
                    if tempParameters[1] < 0:
                        tempParameters[1] = 0
                    if firstEffectFinished:
                        success = applyChorus("copy.wav", "sound.wav", tempParameters[0], tempParameters[1])
                        del tempParameters[0]
                        del tempParameters[0]
                        continue
                    else:
                        success = applyChorus("ctoneshort.wav", "sound.wav", tempParameters[0], tempParameters[1])
                        del tempParameters[0]
                        del tempParameters[0]
                        firstEffectFinished = True
            else:
                testSound(pheno, targetFilepath, fftlength)

    if success:
        distance = getDistance("csd/sound.wav", targetFilepath, fftlength)
        if distance != 0:
            pheno.fitness = distance
        else:
            pheno.fitness = 0.001
    else:
        testSound(pheno, targetFilepath, fftlength)


def testSounds(phenotypes, targetFilepath, fftlength):
    for i in range(len(phenotypes)):
        testSound(phenotypes[i], targetFilepath, fftlength)


# Performs a two point crossover with the genes of two individuals, and creates two children.
# The crossover points are at 1/3 and 2/3 of the vectors, and children inherits order of either mother or father.
def twoPointCrossover(Mother, Father):
    g1 = Genotype()
    g2 = Genotype()

    first = len(Mother.data) / 3
    second = 2 * first

    # Data
    for i in range(0, len(Mother.data)):
        if i >= first and i < second:
            g1.data.append(Mother.data[i])
            g2.data.append(Father.data[i])
        else:
            g1.data.append(Father.data[i])
            g2.data.append(Mother.data[i])

    # Order
    for i in range(len(Mother.order)):
        g1.order.append(Father.order[i])
        g2.order.append(Mother.order[i])

    return g1,g2

# Performs a one point crossover with the genes of two individuals, and creates two children.
# The crossover point is at 1/2 of the bit vector, and children inherits order of either mother or father.
def onePointCrossover(Mother, Father):
    g1 = Genotype()
    g2 = Genotype()

    # Data
    for i in range(0, len(Mother.data)):
        if i < len(Mother.data) / 2:
            g1.data.append(Mother.data[i])
            g2.data.append(Father.data[i])
        else:
            g1.data.append(Father.data[i])
            g2.data.append(Mother.data[i])

    # Order
    for i in range(len(Mother.order)):
        g1.order.append(Father.order[i])
        g2.order.append(Mother.order[i])

    return g1,g2

# Simply clones the mother and father to two children
def clone(Mother, Father):
    g1 = Genotype()
    g2 = Genotype()

    # Data
    for i in range(0, len(Mother.data)):
        g1.data.append(Mother.data[i])
        g2.data.append(Father.data[i])

    # Order
    for i in range(0, len(Mother.order)):
        g1.order.append(Mother.order[i])
        g2.order.append(Father.order[i])


    return g1,g2

# Swaps two bits on a given chance mrate (TYPE 1)
def mutateSwap(bitvector, mrate):
    for i in range(0, len(bitvector)):
        r = random()
        if r <= mrate:
            swap = randint(0,len(bitvector)-1)
            temp = bitvector[swap]
            bitvector[swap] = bitvector[i]
            bitvector[i] = temp

# Performs a mutation on the given bit vector, by flipping the bit on chance mrate ( TYPE 2)
def mutateFlip(bitvector, mrate):
    for i in range(0, len(bitvector)):
        if isinstance(bitvector[i], int):
            a=random()
            if a <= mrate:
                if bitvector[i] == 0:
                    bitvector[i] = 1
                else:
                    bitvector[i] = 0
        if isinstance(bitvector[i], list):
            a=random()
            if a <= mrate:
                for j in range(0, len(bitvector[i])):
                    if bitvector[i][j] == 0:
                        bitvector[i][j] = 1
                    else:
                        bitvector[i][j] = 0

# Performs a mutation on the given order vector, by swapping two effects in the order on chance orate
def mutateOrder(geno, orate):
    order = geno.order
    orderCopy = list(geno.order)
    dataCopy = list(geno.data)
    for i in range(0, len(order)):
        r = random()
        if r <= orate:
            swap = randint(0,len(order)-1)
            if swap != i:
                temp = order[swap]
                order[swap] = order[i]
                order[i] = temp

# Returns the best sound based on fitness, lower is better
def getBestSound(sounds):
    sortedList = sorted(sounds, key=attrgetter('fitness'), reverse=False)
    return sortedList[0]


# Returns the integer value of the input binary vector
def binaryToInteger(data):
    intValue = 0
    for i in range(0, len(data)):
        if data[i] == 1:
            intValue += int(math.pow(2, len(data)-(i+1)))
    return intValue

def average(list):
    return float(sum(list)) / float(len(list))

def std_dev(list):
    mean = float(sum(list)) / float(len(list))
    res = 0.0
    for i in list:
        res += math.pow(i-mean, 2)
    res = res / float(len(list)-1)
    return math.sqrt(res)

def pop_std_dev(list):
    mean = float(sum(list)) / float(len(list))
    res = 0.0
    for i in list:
        res += math.pow(i-mean, 2)
    res = res / float(len(list))
    return math.sqrt(res)

def variance(list):
    return math.pow(pop_std_dev(list), 2)

# This method places the right bits into the parameters (from bitvector to single bit values)
def separateGenes(geno):
    tempData = list(geno.data)
    for j in range(len(geno.order)):
        if geno.order[j] == Enums.DISTORTION:
            for i in range(0,6):
                geno.val_distortion.append(tempData[0])
                del tempData[0]
            continue
        if geno.order[j] == Enums.FLANGER:
            for i in range(0,7):
                geno.val_flanger.append(tempData[0])
                del tempData[0]
            continue
        if geno.order[j] == Enums.TREMOLO:
            for i in range(0,4):
                geno.val_tremolo1.append(tempData[0])
                del tempData[0]
            for i in range(0,4):
                geno.val_tremolo2.append(tempData[0])
                del tempData[0]
            continue
        if geno.order[j] == Enums.CHORUS:
            for i in range(0,4):
                geno.val_chorus1.append(tempData[0])
                del tempData[0]
            for i in range(0,4):
                geno.val_chorus2.append(tempData[0])
                del tempData[0]
            continue

# This method places the parameter bits into the data list (from single bit values)
def distributeGenes(geno):
    del geno.data[:]
    for j in range(len(geno.order)):
        if geno.order[j] == Enums.DISTORTION:
            geno.data.extend(geno.val_distortion)
            continue
        if geno.order[j] == Enums.FLANGER:
            geno.data.extend(geno.val_flanger)
            continue
        if geno.order[j] == Enums.TREMOLO:
            geno.data.extend(geno.val_tremolo1)
            geno.data.extend(geno.val_tremolo2)
            continue
        if geno.order[j] == Enums.CHORUS:
            geno.data.extend(geno.val_chorus1)
            geno.data.extend(geno.val_chorus1)
            continue


# Mates two parents, to create two children.
# Has a given crate chance of cloning the parents to the children
# and a given mrate chance to mutate the bits of the children
def mate(Parents, crossover, crate, mutation, mrate, orderMutation, orate):
    Genotypes = []
    skip = False
    for i in range(0, len(Parents)-1):
        if skip == True:
            skip = False
            continue
        g1 = Genotype()
        g2 = Genotype()
        c = random()
        if c <= crate:
            g1,g2 = clone(Parents[i], Parents[i+1])
            separateGenes(g1)
            separateGenes(g2)
            Genotypes.append(g1)
            Genotypes.append(g2)
            skip = True
            continue
        if(crossover == 1):
            g1,g2 = onePointCrossover(Parents[i], Parents[i+1])
        else:
            g1,g2 = twoPointCrossover(Parents[i], Parents[i+1])
        if mutation == 1:
            mutateSwap(g1.data, mrate)
            mutateSwap(g2.data, mrate)
        else:
            mutateFlip(g1.data, mrate)
            mutateFlip(g2.data, mrate)
        separateGenes(g1)
        separateGenes(g2)
        if orderMutation:
            mutateOrder(g1, orate)
            mutateOrder(g2, orate)
            distributeGenes(g1)
            distributeGenes(g2)
        Genotypes.append(g1)
        Genotypes.append(g2)
        skip = True
    return Genotypes