"""
Selection Mechanisms
--------------------------------------------------------------------------------
This selection mechanism will select partisipants based on their fitness, for
parent selection. These vary in how the partisipants compete; globally or
locally. Also implemented elitism to prevent losing a superior individual.

Containing four mechanisms for parent selection:
- Fitness-proportionate (global)
- Sigma-scaling (global)
- Tournament selection (local)
- Boltzman selection (global)
"""
import random
import math
from operator import attrgetter

#Method to find sum of fitness values of the population
def fitnessSum(adults):
    sum = 0.0
    for i in range(0, len(adults)):
        sum += adults[i].fitness
    return sum

# Method to find average, min and max fitness of the population
def findFitness(adults):
    #Average
    average = fitnessSum(adults)/len(adults)
    #Min
    min = 1000
    for i in range(0, len(adults)):
        if min > adults[i].fitness:
            min = adults[i].fitness
    #Max
    max = 0
    for i in range(0, len(adults)):
        if max > adults[i].fitness:
            max = adults[i].fitness
    return average, min, max

# Method to find average fitness of the total population
def avgTotalFitness(adults):
    length = float(len(adults))
    average = float(fitnessSum(adults)/length)
    return average

# Method to find lowest fitness of the total population
def lowestFitness(adults):
    lowest = 1000
    for i in range(0, len(adults)):
        if lowest > adults[i].fitness:
            lowest = adults[i].fitness
    return lowest

# Method to find highest fitness of the total population
def highestFitness(adults):
    highest = 0
    for i in range(0, len(adults)):
        if highest < adults[i].fitness:
            highest = adults[i].fitness
    return highest

#Method to generate table with the probabilities for each member of the population
#to reproduce
def probTable(adults):
    table = []
    total = fitnessSum(adults)
    for i in range(0, len(adults)):
        table.append(adults[i].fitness / total)
    return table

# Method to find the standard deviation
def stdDeviation(adults):
    stddev = 0.0
    average = avgTotalFitness(adults)
    for i in range(0, len(adults)):
        stddev += math.pow((adults[i].fitness - average), 2.0)
    if len(adults) == 0:
        stddev = 1.0
    else:
        stddev = math.sqrt((stddev/len(adults)))
    return stddev

# Fitness-proprotionate: Fitness scaled so that they sum to 1. No modification
# of relationship with others. Selection based on roulette wheel spin. Takes in
# an array of adults, and their matching probability array
def fitnessProportionate(adults, probtable, amount):
    parents = []
    number = 0

    while(number < amount):
        a = 0
        parent = None
        # rng is a random number between [0, 1> (12 decimals)
        rng = random.random()
        # "Spins" the wheel until the adult with the matching number is found
        for i in range(0, len(adults)):
            a += probtable[i]
            if a >= rng:
                parent = adults[i]
                break
        parents.append(parent)
        number += 1
    return parents

#Sigma-scaling: Modifies the fitness values by using the population's fitness
#variance as a scaling factor.
#Scaling equation: ExpVal(i,g) = 1 + ((f(i)-f_(g))/(2o(g))
#f(i): Fitness of individual i, f_(g): average fitness of generation g
#o(g): Standard deviation of population fitness
def sigmaScaling(adults, amount):
    parents = []
    number = 0
    stddev = stdDeviation(adults)

    # Applys the sigma-scaling to the population
    for i in range(0, len(adults)):
        newfitness = 1.0 + ((adults[i].fitness - avgTotalFitness(adults)) /
        (2.0 * stddev))

        if len(adults) == 0:
            adults[i].fitness = newfitness/1
        else:
            adults[i].fitness = newfitness/fitnessSum(adults)

    probtable = probTable(adults)

    while(number < amount):
        a = 0
        parent = None
        # rng is a random number between [0, 1> (12 decimals)
        rng = random.random()
        # "Spins" the wheel until the adult with the matching number is found
        for i in range(0, len(adults)):
            a += probtable[i]
            if a >= rng:
                parent = adults[i]
                break
        parents.append(parent)
        number += 1
    return parents

# Tournament selection: individuals competition with only a small subset of the
# population; fitness compared and the winner immediately moves to the mating pool.
# amount describes the size of the subset
# E is the probability that the adult with the best fitness is chosen
def tournamentSelection(adults, amount, E):
    parents = []
    number = 0

    while(number < len(adults)):
        subset = []
        for i in range(0, amount):
            rnd = random.randint(0,len(adults)-1)
            subset.append(adults[rnd])
        subset = sorted(subset, key=attrgetter('fitness'), reverse=False)

        rng = random.random()
        if rng <= E:
            parents.append(subset[0])
        if rng > E:
            rnd = random.randint(0,len(subset)-2)
            parents.append(subset[rnd+1])

        number += 1
    return parents

#Boltzmann selection: The degree of randomness is proportianal to a given
#variable (temperature).
#Scaling equation: ExpVal(i,g) = e^f(i)/T)/<e^f(i)/T>g
#g: Generation, T: Temperature, f(i): Fitness of individual i
#<e^f(i)/T>g is the population average of the fitness exponential during g
# Ideally temperature decrease during run
def boltzmannSelection(adults, temperature, amount):
    parents = []
    number = 0
    #avg = avgTotalFitness(adults)
    for i in range(0, len(adults)):
        a = math.exp(adults[i].fitness/temperature)
        #b = math.exp(avg/temperature)
        b = math.exp(avgTotalFitness(adults)/temperature)
        expval = a / b
        normalized = expval/fitnessSum(adults)
        adults[i].fitness = normalized
    probtable = probTable(adults)

    while(number < amount):
        a = 0
        parent = None
        # rng is a random number between [0, 1> (12 decimals)
        rng = random.random()
        # "Spins" the wheel until the adult with the matching number is found
        for i in range(0, len(adults)):
            a += probtable[i]
            if a >= rng:
                parent = adults[i]
                break
        parents.append(parent)
        number += 1
    return parents

